/*  This file is part of the KDE project.

    Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).

    This library is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 2.1 or 3 of the License.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMouseEvent>
#include <QtGui/QPainter>
#include <QPaintEvent>
#include <gst/gst.h>
#include "common.h"
#include "message.h"
#include "mediaobject.h"
#include "qwidgetvideosink.h"
#include "widgetrenderer.h"
#include "qrgb.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <linux/mxcfb.h>

#define MXCFB_GBL_ALPHA 	255
#define MXCFB_CLR_KEY		0x00000000	// ARGB8888

// support old OpenGL installations (1.2)
// assume that if TEXTURE0 isn't defined, none are
#ifndef GL_TEXTURE0
# define GL_TEXTURE0    0x84C0
# define GL_TEXTURE1    0x84C1
# define GL_TEXTURE2    0x84C2
#endif

#ifndef QT_NO_PHONON_VIDEO
QT_BEGIN_NAMESPACE

namespace Phonon
{
namespace Gstreamer
{

WidgetRenderer::WidgetRenderer(VideoWidget *videoWidget)
        : AbstractRenderer(videoWidget)
{
    if ((m_videoSink = gst_element_factory_make("mfw_v4lsink", NULL)) && m_videoSink != NULL) {
	char *videodev;
	gst_object_ref (GST_OBJECT (m_videoSink)); //Take ownership
        gst_object_sink (GST_OBJECT (m_videoSink));
	videodev=getenv("v4lsinkdev");
	if (videodev)
                g_object_set (G_OBJECT (m_videoSink), "device", videodev, NULL);
    }

    // Clear the background with black by default
    QPalette palette;
    palette.setColor(QPalette::Background, Qt::black);
    m_videoWidget->setPalette(palette);
    m_videoWidget->setAutoFillBackground(true);
    m_videoWidget->setAttribute(Qt::WA_NoSystemBackground, false);
    m_videoWidget->setAttribute(Qt::WA_PaintOnScreen, false);
}

WidgetRenderer::~WidgetRenderer()
{
    if (m_videoSink) {
         gst_object_unref (GST_OBJECT (m_videoSink));
         m_videoSink = 0;
     }
}

void WidgetRenderer::setVideoSize(void)
{

   int adj_x;
   int adj_y;	
   
   QSize wSize = m_videoWidget->size();
   m_drawFrameRect = m_videoWidget->calculateDrawFrameRect(); 
   framePos = m_videoWidget->mapToGlobal(QPoint(0,0));

   //Center the video in the widget

   adj_x = (wSize.width()/2) - (m_drawFrameRect.width()/2); 
   adj_y = (wSize.height()/2) - (m_drawFrameRect.height()/2); 
   g_object_set(G_OBJECT(m_videoSink), "axis-left",adj_x + framePos.x(),(const char*)NULL);
   g_object_set(G_OBJECT(m_videoSink), "axis-top", adj_y + framePos.y(), (const char*)NULL);
   g_object_set(G_OBJECT(m_videoSink), "disp-width", m_drawFrameRect.width(), (const char*)NULL);
   g_object_set(G_OBJECT(m_videoSink), "disp-height", m_drawFrameRect.height(), (const char*)NULL);
   g_object_set(G_OBJECT(m_videoSink), "setpara", 1, (const char*)NULL);
}

void WidgetRenderer::handleMediaNodeEvent(const MediaNodeEvent *event)
{
    switch (event->type()) {
    default:
        break;
    }
}

void WidgetRenderer::handlePaint(QPaintEvent *event)
{
    Q_UNUSED(event);
    QPainter painter(m_videoWidget);
    painter.fillRect(m_videoWidget->rect(), m_videoWidget->palette().background());
}

int WidgetRenderer::setOverlay(void)
{
      struct mxcfb_color_key color_key;
      struct mxcfb_gbl_alpha alpha;
      int fd_fb;
	
      if ((fd_fb = open("/dev/fb0", O_RDWR, 0)) < 0)
      {
          printf("Unable to open %s\n", "/dev/fb0");
          return -1;
        
      }
     
      alpha.alpha = MXCFB_GBL_ALPHA;
      alpha.enable = 1;

      if (ioctl(fd_fb, MXCFB_SET_GBL_ALPHA, &alpha) < 0) {
        	printf("Error in applying Alpha\n");
      }

     color_key.color_key = MXCFB_CLR_KEY & 0x00FFFFFF;
     color_key.enable = 1;
     if ( ioctl(fd_fb, MXCFB_SET_CLR_KEY, &color_key) < 0) {
     
	   	printf("Error in applying Color Key\n");
          	return -1;
     }

     close (fd_fb);

     return 0;
}

void WidgetRenderer::handleMove( QMoveEvent * event)
{
     Q_UNUSED(event);

     if (framePos != m_videoWidget->mapToGlobal(QPoint(0,0)))
	setVideoSize();
}

bool WidgetRenderer::eventFilter(QEvent * event)
{
   if (event->type() == QEvent::Show) {
	
	setOverlay();
	return true;

    } else if (event->type() ==  QEvent::Resize) {

	setVideoSize();
	return true;
    }
    if (framePos != m_videoWidget->mapToGlobal(QPoint(0,0)))
	setVideoSize();
    return false;
}

void WidgetRenderer::aspectRatioChanged(Phonon::VideoWidget::AspectRatio)
{
    setVideoSize();
}

void WidgetRenderer::scaleModeChanged(Phonon::VideoWidget::ScaleMode)
{
    setVideoSize();
}
 
void WidgetRenderer::movieSizeChanged(const QSize &movieSize)
{
    Q_UNUSED(movieSize);
    setVideoSize();
}

}
} //namespace Phonon::Gstreamer

QT_END_NAMESPACE
#endif //QT_NO_PHONON_VIDEO
